#include <stdlib.h>
#include <unistd.h>
#include <errno.h>
#include <fcntl.h>
#include <ctype.h>
#include <sys/ioctl.h>
#include <sys/types.h>
#include <linux/i2c-dev.h>

#include "pcf8591.h"

#define PCF8591_ADDR_MIN 0x48 /* PCF8591のアドレスの最小値 */
#define PCF8591_ADDR_MAX 0x4F /* PCF8591のアドレスの最大値 */

#define PCF8591_RETRY_MAX 3 /* read()とwrite()のリトライ回数 */

/* PCF8591のCONTROL BYTEの設定 */
#define PCF8591_ANALOG_OUTPUT_ENABLE          (1 << 6)
#define PCF8591_ANALOG_OUTPUT_DISABLE         (0 << 6)

#define PCF8591_SINGLE_ENDED                  (0 << 4)
#define PCF8591_THREE_DIFFERENTIAL            (1 << 4)
#define PCF8591_SINGLE_ENDED_AND_DIFFERENTIAL (2 << 4)
#define PCF8591_TWO_DIFFERENTIAL              (3 << 4)

#define PCF8591_AUTO_INCREMENT_ENABLE         (1 << 2)
#define PCF8591_AUTO_INCREMENT_DISABLE        (0 << 2)

#define PCF8591_CH_SHIFT                      (0)


struct pcf8591 {
	int fd;
};

static ssize_t write_uninterruptible(int fd, const void *buf, size_t count,
				     int retry)
{
	ssize_t ret;
	int i;

	for (i = 0; i < retry; i++) {
		ret = write(fd, buf, count);
		if (ret < 0 && errno != EINTR)
			return -1;

		if ((size_t)ret == count)
			return ret;
	}

	errno = ETIMEDOUT;
	return -1;
}

static ssize_t read_uninterruptible(int fd, void *buf, size_t count, int retry)
{
	size_t read_length = 0;
	ssize_t ret;
	int i;

	for (i = 0; i < retry; i++) {
		ret = read(fd, buf + read_length, count - read_length);
		if (ret < 0) {
			if (errno == EINTR)
				continue;
			else
				return -1;
		}

		read_length += ret;
		if (read_length == count)
			return read_length;
	}

	errno = ETIMEDOUT;
	return -1;
}

/**
 * 指定されたデバイスファイルをオープンする
 *
 * @param dev_path  PCF8591が接続されたi2cdevのデバイスファイルへのパス。
 * @param addr      PCF8591のアドレス。
 *
 * @return 成功するとstruct pcf8591へのポインタを返す。
 *         失敗するとNULLを返す。その際、適切なerrnoを設定する。
 */
struct pcf8591 *pcf8591_open(const char *dev_path, const int addr)
{
	struct pcf8591 *adc;
	int error;
	int ret;

	if (dev_path == NULL ||
	    addr < PCF8591_ADDR_MIN || PCF8591_ADDR_MAX < addr) {
		errno = EINVAL;
		return NULL;
	}

	adc = calloc(1, sizeof(struct pcf8591));
	if (adc == NULL)
		return NULL;

	adc->fd = open(dev_path, O_RDWR);
	if (adc->fd < 0) {
		error = errno;
		goto err1;
	}

	ret = ioctl(adc->fd, I2C_SLAVE, addr);
	if (ret < 0) {
		error = errno;
		goto err2;
	}

	return adc;

err2:
	close(adc->fd);
err1:
	free(adc);
	errno = error;
	return NULL;
}

/**
 * 指定されたチャンネルの値をA/Dコンバータにセットし、
 * PCF8591からA/D変換結果を読み込む。
 * 読み込んだA/D変換結果をdigitに格納する。
 *
 * @param adc    オープン済みのPCF8591デバイスへのポインタ。
 * @param ch     サンプリングするチャンネル。
 * @param digit  A/D変換された値が格納される。
 *
 * @return 成功すると0を返し、digitにA/D変換された値を格納する。
 *         失敗すると-1を返す。その際、適切なerrnoを設定する。
 */
int pcf8591_read(struct pcf8591 *adc, const int ch, uint8_t *digit)
{
	uint8_t buf[2];
	int ret;

	if (adc == NULL || digit == NULL ||
	    ch < PCF8591_CH_MIN || PCF8591_CH_MAX < ch) {
		errno = EINVAL;
		return -1;
	}

	buf[0] = PCF8591_ANALOG_OUTPUT_DISABLE |
	         PCF8591_SINGLE_ENDED |
	         PCF8591_AUTO_INCREMENT_DISABLE |
	         (ch << PCF8591_CH_SHIFT);

	ret = write_uninterruptible(adc->fd, buf, 1, PCF8591_RETRY_MAX);
	if (ret < 0)
		return -1;

	/* 現在のアナログ入力を変換した値を取得するために、2バイト読み込む */
	ret = read_uninterruptible(adc->fd, buf, 2, PCF8591_RETRY_MAX);
	if (ret < 0)
		return -1;

	*digit = buf[1];

	return 0;
}

/**
 * 指定されたPCF8591デバイスをクローズする
 *
 * @param adc オープン済みのPCF8591デバイスへのポインタ。
 *
 * @return 成功すると0を返す。
 *         失敗すると-1を返す。その際、適切なerrnoを設定する。
 */
int pcf8591_close(struct pcf8591 *adc)
{
	int fd;

	if (adc == NULL) {
		errno = EINVAL;
		return -1;
	}

	fd = adc->fd;
	free(adc);

	return close(fd);
}
